
/******************************************************************************
* MODULE     : list.hpp
* DESCRIPTION: Linked lists
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_LIST_HPP
#define __MML_LIST_HPP
#include "iterator.hpp"
namespace mml {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>
#define List list<C>
#define List_rep list_rep<C>
TMPL class list_rep;
TMPL class list;
TMPL List cons (const C& c, const List& l);
TMPL C car (const List& l);
TMPL C& ref_car (List& l);
TMPL List cdr (const List& l);
TMPL List& ref_cdr (List& l);
TMPL bool nil (const List& l);

/******************************************************************************
* List class and its representation class
******************************************************************************/

TMPL_DEF
class list_rep REP_STRUCT {
  C    item;
  List next;

public:
  List_rep (const C& item2, const List& next2):
    item(item2), next(next2) {}
  friend class List;
  friend C car LESSGTR (const List& l);
  friend C& ref_car LESSGTR (List& l);
  friend List cdr LESSGTR (const List& l);
  friend List& ref_cdr LESSGTR (List& l);
};

TMPL_DEF
class list {
INDIRECT_PROTO_1 (list, list_rep, C)
private:
  list (const C& item, const List& next):
    rep (new List_rep (item, next)) {}
public:
  list (): rep (NULL) {}
  list (const C& c1):
    rep (new List_rep (c1, List ())) {}
  list (const C& c1, const C& c2):
    rep (new List_rep (c1, List (c2))) {}
  list (const C& c1, const C& c2, const C& c3):
    rep (new List_rep (c1, List (c2, c3))) {}
  list (const C& c1, const C& c2, const C& c3, const C& c4):
    rep (new List_rep (c1, List (c2, c3, c4))) {}
  list (const iterator<C>& it) {
    if (busy (it)) { C c= *it; ++it; rep= new List_rep (c, it); }
    else rep= NULL; }
  C& operator () (nat i) {
    assert (rep != NULL);
    secure ();
    return i==0? rep->item: rep->next[i-1]; }
  C  operator [] (nat i) const {
    assert (rep != NULL);
    return i==0? rep->item: rep->next[i-1]; }
  friend List cons LESSGTR (const C& c, const List& l);
  friend C car LESSGTR (const List& l);
  friend C& ref_car LESSGTR (List& l);
  friend List cdr LESSGTR (const List& l);
  friend List& ref_cdr LESSGTR (List& l);
  friend bool nil LESSGTR (const List& l);
};
INDIRECT_IMPL_1 (list, list_rep, typename C, C)

/******************************************************************************
* Basic constructors and accessors
******************************************************************************/

TMPL List cons (const C& c, const List& l) { return List (c, l); }
TMPL C car (const List& l) { return l.rep->item; }
TMPL C& ref_car (List& l) { l.secure(); return l.rep->item; }
TMPL List cdr (const List& l) { return l.rep->next; }
TMPL List& ref_cdr (List& l) { l.secure(); return l.rep->next; }
TMPL bool nil (const List& l) { return l.rep == NULL; }
TMPL int N (const List& l) { return nil (l)? 0: N (cdr (l)) + 1; }
TMPL bool atom (const List& l) { return N(l)==1; }

/******************************************************************************
* List iteration
******************************************************************************/

TMPL_DEF
class list_iterator_rep: public iterator_rep<C> {
  List l;  // the list to traverse
protected:
  bool is_busy () { return !nil (l); }
  void advance () { l= cdr (l); }
  C current () { return car (l); }
public:
  list_iterator_rep (const List& l2): l(l2) {}
};

TMPL iterator<C>
iterate (const List& l) {
  return iterator<C> (new list_iterator_rep<C> (l));
}

/******************************************************************************
* Printing lists
******************************************************************************/

TMPL content
as_content (const List& l) {
  content args;
  List counter= l;
  while (!nil (counter)) {
    args << as_content (car (counter));
    counter= cdr (counter);
  }
  return apply ("list", args);
}

PRINT_SUGAR(TMPL,List)

/******************************************************************************
* Operations on lists
******************************************************************************/

TMPL List
copy (const List& l) {
  if (nil (l)) return l;
  return cons (car (l), cdr (l));
}

TMPL List
operator * (const List& l1, const List& l2) {
  if (nil (l1)) return copy (l2);
  return cons (car (l1), cdr (l1) * l2);
}

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL bool
operator == (const List& l1, const List& l2) {
  if (nil (l1)) return nil (l2);
  return car (l1) == car (l2) && cdr (l1) == cdr (l2);
}

TMPL bool
operator != (const List& l1, const List& l2) {
  if (nil (l1)) return !nil (l2);
  return car (l1) != car (l2) || cdr (l1) != cdr (l2);
}

#undef TMPL_DEF
#undef TMPL
#undef List
#undef List_rep
} // namespace mml
#endif // __MML_LIST_HPP
