
/******************************************************************************
* MODULE     : symbolic_pretty.cpp
* DESCRIPTION: pretty printer for symbolic expressions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "symbolic.hpp"
#include "list.hpp"
namespace mml {

/******************************************************************************
* The pretty printer class
******************************************************************************/

class pretty_printer {
private:
  ostream&       out;
  list<symbolic> sl;

private:
  void pp_verb (const string& s);
  void pp_op (const symbolic& op, bool infix= true);
  void pp_keyw (bool before, const string& s, bool after);

  void pp_E0s (const symbolic& s, nat start);
  void pp_Vtuple (const symbolic& s, nat start);

  void pp_E1 (const symbolic& g);
  void pp_Lflux (const symbolic& g);
  void pp_Rflux (const symbolic& g);

  void pp_F0 (const symbolic& s);
  void pp_Ors (const symbolic& s);
  void pp_Xors (const symbolic& s);
  void pp_F1 (const symbolic& s);
  void pp_Ands (const symbolic& s);
  void pp_F2 (const symbolic& s);
  void pp_F3 (const symbolic& s);

  void pp_T0 (const symbolic& s);
  void pp_T1 (const symbolic& s);
  void pp_T2 (const symbolic& s);
  void pp_T3 (const symbolic& s);
  void pp_T4 (const symbolic& s);
  void pp_T5 (const symbolic& s);

  void pp_R0 (const symbolic& s);
  void pp_R1 (const symbolic& s);
  void pp_R2 (const symbolic& s);
  void pp_R3 (const symbolic& s);

public:
  inline pretty_printer (ostream& out2): out (out2), sl () {}
  inline ~pretty_printer () {}

  void pp_E0 (const symbolic& s);
};

/******************************************************************************
* Helper routines
******************************************************************************/

bool
is_func (const symbolic& s, const symbolic& f) {
  return !is_atomic (s) && s[0] == f;
}

bool
is_func (const symbolic& s, const symbolic& f, nat n) {
  return !is_atomic (s) && N(s) == n && s[0] == f;
}

void
pretty_printer::pp_verb (const string& s) {
  out << s;
}

void
pretty_printer::pp_op (const symbolic& op, bool infix) {
  if (infix) out << " ";
  out << L(op);
  if (infix) out << " ";
}

void
pretty_printer::pp_keyw (bool before, const string& s, bool after) {
  if (before) out << " ";
  out << s;
  if (after) out << " ";
}

/******************************************************************************
* Tuples and vertical tuples
******************************************************************************/

void
pretty_printer::pp_E0s (const symbolic& s, nat start) {
  nat n= N(s);
  for (nat i=start; i<=n; i++) {
    if (i != start) pp_verb (", ");
    pp_E0 (s[i]);
  }
}

void
pretty_printer::pp_Vtuple (const symbolic& s, nat start) {
  nat n= N(s);
  for (nat i=start; i<=n; i++) {
    if (i != start) pp_verb (";\n");
    if (is_func (s[i], SYMB_TUPLE)) pp_E0s (s[i], 1);
    else pp_E0 (s[i]);
  }
}

/******************************************************************************
* Expressions
******************************************************************************/

void
pretty_printer::pp_E0 (const symbolic& s) {
  pp_E1 (s);
}

void
pretty_printer::pp_E1 (const symbolic& s) {
  if (is_func (s, SYMB_LESSLESS, 2)) {
    pp_Lflux (s[1]);
    pp_op (s[0]);
    pp_F0 (s[2]);
  }
  else if (is_func (s, SYMB_GTRGTR, 2)) {
    pp_F0 (s[1]);
    pp_op (s[0]);
    pp_Rflux (s[2]);
  }
  else pp_F0 (s);
}

void
pretty_printer::pp_Lflux (const symbolic& s) {
  if (is_func (s, SYMB_LESSLESS, 2)) {
    pp_Lflux (s[1]);
    pp_op (s[0]);
    pp_F0 (s[2]);
  }
  else pp_F0 (s);
}

void
pretty_printer::pp_Rflux (const symbolic& s) {
  if (is_func (s, SYMB_GTRGTR, 2)) {
    pp_F0 (s[1]);
    pp_op (s[0]);
    pp_Rflux (s[2]);
  }
  else pp_F0 (s);
}

/******************************************************************************
* Formulas
******************************************************************************/

void
pretty_printer::pp_F0 (const symbolic& s) {
  if (is_func (s, SYMB_IMPLIES, 2) ||
      is_func (s, SYMB_EQUIV, 2)) {
    pp_F1 (s[1]);
    pp_op (s[0]);
    pp_F1 (s[2]);
  }
  else pp_F1 (s);
}

void
pretty_printer::pp_F1 (const symbolic& s) {
  if (is_func (s, SYMB_OR, 2)) {
    pp_Xors (s[1]);
    pp_op (s[0]);
    pp_Xors (s[2]);
  }
  else if (is_func (s, SYMB_XOR, 2)) {
    pp_Xors (s[1]);
    pp_op (s[0]);
    pp_Xors (s[2]);
  }
  else pp_F2 (s);
}

void
pretty_printer::pp_Ors (const symbolic& s) {
  if (is_func (s, SYMB_OR, 2)) {
    pp_Ors (s[1]);
    pp_op (s[0]);
    pp_Ors (s[2]);
  }
  else pp_F2 (s);
}

void
pretty_printer::pp_Xors (const symbolic& s) {
  if (is_func (s, SYMB_XOR, 2)) {
    pp_Xors (s[1]);
    pp_op (s[0]);
    pp_Xors (s[2]);
  }
  else pp_F2 (s);
}

void
pretty_printer::pp_F2 (const symbolic& s) {
  if (is_func (s, SYMB_AND, 2)) {
    pp_Ands (s[1]);
    pp_op (s[0]);
    pp_Ands (s[2]);
  }
  else pp_F3 (s);
}

void
pretty_printer::pp_Ands (const symbolic& s) {
  if (is_func (s, SYMB_AND, 2)) {
    pp_Ands (s[1]);
    pp_op (s[0]);
    pp_Ands (s[2]);
  }
  else pp_F3 (s);
}

void
pretty_printer::pp_F3 (const symbolic& s) {
  if (is_func (s, SYMB_EQUAL, 2) ||
      is_func (s, SYMB_UNEQUAL, 2) ||
      is_func (s, SYMB_LESS, 2) ||
      is_func (s, SYMB_LESSEQ, 2) ||
      is_func (s, SYMB_GTR, 2) ||
      is_func (s, SYMB_GTREQ, 2)) {
    pp_T0 (s[1]);
    pp_op (s[0]);
    pp_T0 (s[2]);
  }
  else pp_T0 (s);
}

/******************************************************************************
* Terms
******************************************************************************/

void
pretty_printer::pp_T0 (const symbolic& s) {
  pp_T1 (s);
}

void
pretty_printer::pp_T1 (const symbolic& s) {
  if (is_func (s, SYMB_FUNCTION, 2)) {
    pp_T2 (s[1]);
    pp_op (s[0]);
    pp_T2 (s[2]);
  }
  else pp_T2 (s);
}

void
pretty_printer::pp_T2 (const symbolic& s) {
  pp_T3 (s);
}

void
pretty_printer::pp_T3 (const symbolic& s) {
  if (is_func (s, SYMB_PLUS, 2)) {
    pp_T3 (s[1]);
    pp_op (s[0]);
    pp_T3 (s[2]);
  }
  else if (is_func (s, SYMB_MINUS, 2)) {
    pp_T3 (s[1]);
    pp_op (s[0]);
    pp_T4 (s[2]);
  }
  else pp_T4 (s);
}

void
pretty_printer::pp_T4 (const symbolic& s) {
  if (is_func (s, SYMB_TIMES, 2) ||
      is_func (s, SYMB_OVER, 2) ||
      is_func (s, SYMB_COMPOSE, 2)) {
    pp_T4 (s[1]);
    pp_op (s[0]);
    pp_T4 (s[2]);
  }
  else pp_T5 (s);
}

void
pretty_printer::pp_T5 (const symbolic& s) {
  if (is_func (s, SYMB_NOT, 1) ||
      is_func (s, SYMB_MINUS, 1)) {
    pp_op (s[0], false);
    pp_R0 (s[1]);
  }
  else pp_R0 (s);
}

/******************************************************************************
* Radicals
******************************************************************************/

void
pretty_printer::pp_R0 (const symbolic& s) {
  pp_R1 (s);
}

void
pretty_printer::pp_R1 (const symbolic& s) {
  if (is_func (s, SYMB_POWER, 2)) {
    pp_R2 (s[1]);
    pp_op (s[0], false); // suppress spaces
    pp_R2 (s[2]);
  }
  else pp_R2 (s);
}

static bool
is_applicable (const symbolic& x) {
  if (is_atomic (x)) {
    string s= L(x);
    if (N(s)==0) return false;
    for (nat i=0; i<N(s); i++)
      if (((s[i]<'0') || (s[i]>'9')) &&
          ((s[i]<'a') || (s[i]>'z')) &&
          ((s[i]<'A') || (s[i]>'Z')) &&
          ((s[i]<'_') || (s[i]>'_')))
        return false;
  }
  return true;
}

void
pretty_printer::pp_R2 (const symbolic& s) {
  if (is_func (s, SYMB_FACTORIAL, 1)) {
    pp_R2 (s[1]);
    pp_verb ("!");
  }
  else if (is_func (s, SYMB_PRIME, 1)) {
    pp_R2 (s[1]);
    pp_op (s[0], false);
  }
  else if (is_func (s, SYMB_DOT, 2)) {
    pp_R2 (s[1]);
    pp_verb (".");
    pp_R3 (s[2]);
  }
  else if (is_func (s, SYMB_ACCESS) && (N(s)>1)) {
    pp_R2 (s[1]);
    pp_verb ("[");
    if ((N(s) == 2) && is_func (s[2], SYMB_VTUPLE))
      pp_Vtuple (s[2], 1);
    else pp_E0s (s, 2);
    pp_verb ("]");
  }
  else if ((!is_atomic (s) && is_applicable (s[0])) ||
	   (!nil (sl)) && (car (sl) == s)) {
    pp_R2  (s[0]);
    pp_verb (" (");
    pp_E0s (s, 1);
    pp_verb (")");
  }
  else pp_R3 (s);
}

void
pretty_printer::pp_R3 (const symbolic& s) {
  if (is_atomic (s))
    pp_verb (L(s));
  else if (is_func (s, SYMB_SQTUPLE)) {
    pp_verb ("[");
    if ((N(s) == 1) && is_func (s[1], SYMB_VTUPLE))
      pp_Vtuple (s[1], 1);
    else pp_E0s (s, 1);
    pp_verb ("]");
  }
  else if (is_func (s, SYMB_TUPLE)) {
    pp_verb ("(");
    pp_E0s (s, 1);
    pp_verb (")");
  }
  else if (is_func (s, SYMB_VTUPLE)) {
    pp_verb ("(");
    pp_Vtuple (s, 1);
    pp_verb (")");
  }
  else {
    pp_verb ("(");
    sl= cons (s, sl);
    pp_E0 (s);
    sl= cdr (sl);
    pp_verb (")");
  }
}

/******************************************************************************
* Interface
******************************************************************************/

void
pretty_print (ostream& out, const symbolic& s) {
  pretty_printer pp (out);
  symbolic t= flatten (s);
  pp.pp_E0 (t);
}

} // namespace mml
