
/******************************************************************************
* MODULE     : dense_vector.hpp
* DESCRIPTION: Dense vectors
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_DENSE_VECTOR_HPP
#define __MML_DENSE_VECTOR_HPP
#include "vector.hpp"
#include "segment.hpp"
namespace mml {
#define TMPL template<typename C, typename V>
#define Vector vector<C,dense<V> >
#define Vector_rep vector_rep<C,dense<V> >
TMPL nat dim (const Vector& v);
TMPL C* seg (Vector& v);
TMPL const C* seg (const Vector& v);

/******************************************************************************
* Dense vectors of arbitrary sizes
******************************************************************************/

template<typename C, typename V>
class Vector_rep REP_STRUCT {
  C*  a;  // entries of vector
  nat n;  // dimension of vector
  nat l;  // allocated number of entries
public:
  vector_rep (C* a2, nat n2): a (a2), n (n2), l (n2) {}
  ~vector_rep () { delete[] a; }
  void resize (nat n2) {
    if (n2 > l) {
      nat l2= max (n2, l<<1);
      C* b= new C[l2];
      segment_copy (b, a, n);
      delete[] a;
      a= b;
      l= l2;
    }
    n= n2;
  }
  friend class Vector;
  friend nat dim LESSGTR (const Vector& v);
  friend C* seg LESSGTR (Vector& v);
  friend const C* seg LESSGTR (const Vector& v);
};

template<typename C, typename V>
class Vector {
INDIRECT_PROTO_2 (vector, vector_rep, C, dense<V>)
public:
  vector () {
    nat n= dense<V>::def_len;
    rep= new Vector_rep (new C[n], n);
  }
  vector (int x) {
    nat n= dense<V>::def_len;
    rep= new Vector_rep (new C[n], n);
    segment_set (rep->a, C(x), n);
  }
  vector (const C& x, nat n=dense<V>::def_len) {
    rep= new Vector_rep (new C[n], n);
    segment_set (rep->a, x, n);
  }
  vector (C* a, nat n=dense<V>::def_len) {
    rep= new Vector_rep (a, n);
  }
  vector (const iterator<C>& it) {
    nat i, n= dense<V>::def_len;
    if (n==0) n=1;
    C* a= new C[n];
    for (i=0; busy (it); i++, ++it) {
      if (i >= n) {
	C* b= new C[n<<1];
	segment_copy (b, a, n);
	delete[] a;
	a= b;
	n= n<<1;
      }
      a[i]= *it;
    }
    rep= new Vector_rep (a, i);
  }
  C& operator () (nat i) { secure(); return rep->a[i]; }
  C operator [] (nat i) const { return rep->a[i]; }
  Vector& operator << (const C& append);
  Vector& operator << (const Vector& w);
};
INDIRECT_IMPL_2 (vector, vector_rep, typename C, C, typename V, dense<V>)

TMPL nat dim (const Vector& v) { return v->n; }
TMPL C* seg (Vector& v) { v.secure(); return v->a; }
TMPL const C* seg (const Vector& v) { return v->a; }

/******************************************************************************
* Fixed size dense vectors
******************************************************************************/

template<typename C, nat N>
class vector_rep<C,dense<linear<naive,N> > > REP_STRUCT {
  C* a;
public:
  vector_rep (C* a2, nat dummy): a (a2) { (void) dummy; }
  virtual ~vector_rep () { delete[] a; }
  friend class vector<C,dense<linear<naive,N> > >;
  friend nat dim LESSGTR (const vector<C,dense<linear<naive,N> > >& v);
  friend C*  seg LESSGTR (vector<C,dense<linear<naive,N> > >& v);
  friend const C* seg LESSGTR (const vector<C,dense<linear<naive,N> > >& v);
};

template<typename C, nat N> nat
dim (const vector<C,dense<linear<naive,N> > >& v) { return N; }

/******************************************************************************
* Explicit dense vectors
******************************************************************************/

template<typename C> vector<C>
vec () {
  C* a= new C[0];
  return vector<C> (a, 0);
}

template<typename C> vector<C>
vec (const C& c1) {
  C* a= new C[1];
  a[0]= c1;
  return vector<C> (a, 1);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2) {
  C* a= new C[2];
  a[0]= c1; a[1]= c2;
  return vector<C> (a, 2);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3) {
  C* a= new C[3];
  a[0]= c1; a[1]= c2; a[2]= c3;
  return vector<C> (a, 3);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3, const C& c4) {
  C* a= new C[4];
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4;
  return vector<C> (a, 4);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3, const C& c4, const C& c5) {
  C* a= new C[5];
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4; a[4]= c5;
  return vector<C> (a, 5);
}

/******************************************************************************
* Appending new elements to vectors
******************************************************************************/

TMPL Vector
copy (const Vector& v) {
  nat n= dim (v);
  C* a= new C[n];
  segment_copy (a, seg (v), n);
  return Vector (a, n);
}

TMPL Vector
append (const Vector& v, const Vector& w) {
  nat n= dim (v), p= dim (w);
  C* a= new C[n+p];
  segment_copy (a, seg (v), n);
  segment_copy (a+n, seg (w), p);
  return Vector (a, n+p);
}

TMPL Vector&
Vector::operator << (const C& append) {
  secure ();
  nat n= rep->n;
  rep->resize (n+1);
  rep->a[n]= append;
  return *this;
}

TMPL Vector&
Vector::operator << (const Vector& w) {
  secure ();
  nat n= rep->n, p= dim (w);
  rep->resize (n + p);
  segment_copy (rep->a + n, seg (w), p);
  return *this;
}

/******************************************************************************
* Routines on dense vectors
******************************************************************************/

#define VECTOR_UNARY(name,op)			\
TMPL Vector					\
operator op (const Vector& x) {			\
  nat n= dim (x);				\
  C* r= new C[n];				\
  name<C> (r, seg (x), n);			\
  return Vector (r, n);				\
}

#define VECTOR_BINARY(name,op)				\
TMPL Vector						\
operator op (const Vector& x, const Vector& y) {	\
  nat n= dim (x);					\
  assert (dim (y) == n);				\
  C* r= new C[n];					\
  name<C> (r, seg (x), seg (y), n);			\
  return Vector (r, n);					\
}

#define VECTOR_SCALAR(name,op)			\
TMPL Vector					\
operator op (const Vector& x, const C c) {	\
  nat n= dim (x);				\
  C* r= new C[n];				\
  name<C> (r, seg (x), c, n);			\
  return Vector (r, n);				\
}

#define VECTOR_VAR_SCALAR(name,op)		\
TMPL Vector					\
operator op (const C c, const Vector& x) {	\
  nat n= dim (x);				\
  C* r= new C[n];				\
  name<C> (r, seg (x), c, n);			\
  return Vector (r, n);				\
}

#define VECTOR_INLINE_BINARY(name,op)		\
TMPL Vector&					\
operator op (Vector& x, const Vector& y) {	\
  nat n= dim (x);				\
  assert (dim (y) == n);			\
  name<C> (seg (x), seg (y), n);		\
  return x;					\
}  

#define VECTOR_INLINE_SCALAR(name,op)		\
TMPL Vector&					\
operator op (Vector& x, const C& c) {		\
  nat n= dim (x);				\
  name<C> (seg (x), c, n);			\
  return x;					\
}  

#define VECTOR_TEST(name,op)				\
TMPL bool						\
operator op (const Vector& x, const Vector& y) {	\
  nat n= dim (x);					\
  assert (dim (y) == n);				\
  return name<C> (seg (x), seg (y), n);			\
}

TMPL bool
operator == (const Vector& x, const Vector& y) {
  nat n= dim (x);
  if (dim (y) != n) return false;
  return segment_eq<C> (seg (x), seg (y), n);
}

VECTOR_UNARY(segment_neg,-);
VECTOR_BINARY(segment_add,+);
VECTOR_BINARY(segment_sub,-);
VECTOR_BINARY(segment_mul,*);
VECTOR_BINARY(segment_div,/);
VECTOR_BINARY(segment_mod,%);
VECTOR_SCALAR(segment_mul,*);
VECTOR_VAR_SCALAR(segment_mul,*);
VECTOR_SCALAR(segment_div,/);
VECTOR_SCALAR(segment_mod,%);
VECTOR_INLINE_BINARY(segment_add,+=);
VECTOR_INLINE_BINARY(segment_sub,-=);
VECTOR_INLINE_BINARY(segment_mul,*=);
VECTOR_INLINE_BINARY(segment_div,/=);
VECTOR_INLINE_BINARY(segment_mod,%=);
VECTOR_INLINE_SCALAR(segment_mul,*=);
VECTOR_INLINE_SCALAR(segment_div,/=);
VECTOR_INLINE_SCALAR(segment_mod,%=);
VECTOR_TEST(segment_lesseq,<=);
VECTOR_TEST(segment_gtreq,>=);

ARITH_INT_SUGAR(TMPL,Vector,C);
STRICT_COMPARE_SUGAR(TMPL,Vector)

#undef TMPL
#undef Vector
#undef Vector_rep
} // namespace mml
#endif // __MML_DENSE_VECTOR_HPP
